@@ -20,8 +20,10 @@ from .. import mlog
 from .. import compilers
 import json
 import subprocess
-from ..mesonlib import MesonException, get_compiler_for_source, classify_unity_sources
+from ..mesonlib import MesonException, get_meson_script
+from ..mesonlib import get_compiler_for_source, classify_unity_sources
 from ..compilers import CompilerArgs
+from collections import OrderedDict
 
 class CleanTrees:
     '''
@@ -33,10 +35,11 @@ class CleanTrees:
         self.trees = trees
 
 class InstallData:
-    def __init__(self, source_dir, build_dir, prefix):
+    def __init__(self, source_dir, build_dir, prefix, strip_bin, mesonintrospect):
         self.source_dir = source_dir
         self.build_dir = build_dir
         self.prefix = prefix
+        self.strip_bin = strip_bin
         self.targets = []
         self.headers = []
         self.man = []
@@ -45,6 +48,7 @@ class InstallData:
         self.po = []
         self.install_scripts = []
         self.install_subdirs = []
+        self.mesonintrospect = mesonintrospect
 
 class ExecutableSerialisation:
     def __init__(self, name, fname, cmd_args, env, is_cross, exe_wrapper,
@@ -75,6 +79,24 @@ class TestSerialisation:
         self.workdir = workdir
         self.extra_paths = extra_paths
 
+class OptionProxy:
+    def __init__(self, name, value):
+        self.name = name
+        self.value = value
+
+class OptionOverrideProxy:
+    '''Mimic an option list but transparently override
+    selected option values.'''
+    def __init__(self, overrides, options):
+        self.overrides = overrides
+        self.options = options
+
+    def __getitem__(self, option_name):
+        base_opt = self.options[option_name]
+        if option_name in self.overrides:
+            return OptionProxy(base_opt.name, base_opt.validate_value(self.overrides[option_name]))
+        return base_opt
+
 # This class contains the basic functionality that is needed by all backends.
 # Feel free to move stuff in and out of it as you see fit.
 class Backend:
@@ -102,6 +124,12 @@ class Backend:
     def get_target_filename_abs(self, target):
         return os.path.join(self.environment.get_build_dir(), self.get_target_filename(target))
 
+    def get_option_for_target(self, option_name, target):
+        if option_name in target.option_overrides:
+            override = target.option_overrides[option_name]
+            return self.environment.coredata.validate_option_value(option_name, override)
+        return self.environment.coredata.get_builtin_option(option_name)
+
     def get_target_filename_for_linking(self, target):
         # On some platforms (msvc for instance), the file that is used for
         # dynamic linking is not the same as the dynamic library itself. This
@@ -114,12 +142,6 @@ class Backend:
             return os.path.join(self.get_target_dir(target), target.get_filename())
         raise AssertionError('BUG: Tried to link to something that\'s not a library')
 
-    def get_target_debug_filename(self, target):
-        fname = target.get_debug_filename()
-        if not fname:
-            raise AssertionError("BUG: Tried to generate debug filename when it doesn't exist")
-        return os.path.join(self.get_target_dir(target), fname)
-
     def get_target_dir(self, target):
         if self.environment.coredata.get_builtin_option('layout') == 'mirror':
             dirname = target.get_subdir()
@@ -157,8 +179,10 @@ class Backend:
         compsrcs = classify_unity_sources(target.compilers.values(), unity_src)
 
         def init_language_file(suffix):
-            outfilename = os.path.join(self.get_target_private_dir_abs(target),
-                                       self.get_unity_source_filename(target, suffix))
+            unity_src_name = self.get_unity_source_filename(target, suffix)
+            unity_src_subdir = self.get_target_private_dir_abs(target)
+            outfilename = os.path.join(unity_src_subdir,
+                                       unity_src_name)
             outfileabs = os.path.join(self.environment.get_build_dir(),
                                       outfilename)
             outfileabs_tmp = outfileabs + '.tmp'
@@ -166,7 +190,7 @@ class Backend:
             outfileabs_tmp_dir = os.path.dirname(outfileabs_tmp)
             if not os.path.exists(outfileabs_tmp_dir):
                 os.makedirs(outfileabs_tmp_dir)
-            result.append(outfilename)
+            result.append(mesonlib.File(True, unity_src_subdir, unity_src_name))
             return open(outfileabs_tmp, 'w')
 
         # For each language, generate a unity source file and return the list
@@ -191,7 +215,7 @@ class Backend:
             elif isinstance(obj, mesonlib.File):
                 obj_list.append(obj.rel_to_builddir(self.build_to_src))
             elif isinstance(obj, build.ExtractedObjects):
-                obj_list += self.determine_ext_objs(obj, proj_dir_to_build_root)
+                obj_list += self.determine_ext_objs(target, obj, proj_dir_to_build_root)
             else:
                 raise MesonException('Unknown data type in object list.')
         return obj_list
@@ -215,13 +239,13 @@ class Backend:
         exe_data = os.path.join(self.environment.get_scratch_dir(), scratch_file)
         with open(exe_data, 'wb') as f:
             if isinstance(exe, dependencies.ExternalProgram):
-                exe_fullpath = exe.fullpath
+                exe_cmd = exe.get_command()
                 exe_needs_wrapper = False
             elif isinstance(exe, (build.BuildTarget, build.CustomTarget)):
-                exe_fullpath = [self.get_target_filename_abs(exe)]
+                exe_cmd = [self.get_target_filename_abs(exe)]
                 exe_needs_wrapper = exe.is_cross
             else:
-                exe_fullpath = [exe]
+                exe_cmd = [exe]
                 exe_needs_wrapper = False
             is_cross = exe_needs_wrapper and \
                 self.environment.is_cross_build() and \
@@ -231,11 +255,11 @@ class Backend:
                 exe_wrapper = self.environment.cross_info.config['binaries'].get('exe_wrapper', None)
             else:
                 exe_wrapper = None
-            if mesonlib.is_windows():
+            if mesonlib.is_windows() or mesonlib.is_cygwin():
                 extra_paths = self.determine_windows_extra_paths(exe)
             else:
                 extra_paths = []
-            es = ExecutableSerialisation(basename, exe_fullpath, cmd_args, env,
+            es = ExecutableSerialisation(basename, exe_cmd, cmd_args, env,
                                          is_cross, exe_wrapper, workdir,
                                          extra_paths, capture)
             pickle.dump(es, f)
@@ -266,32 +290,34 @@ class Backend:
             raise MesonException(m.format(target.name))
         return l
 
-    def object_filename_from_source(self, target, source):
+    def object_filename_from_source(self, target, source, is_unity):
         if isinstance(source, mesonlib.File):
             source = source.fname
         # foo.vala files compile down to foo.c and then foo.c.o, not foo.vala.o
         if source.endswith('.vala'):
+            if is_unity:
+                return source[:-5] + '.c.' + self.environment.get_object_suffix()
             source = os.path.join(self.get_target_private_dir(target), source[:-5] + '.c')
         return source.replace('/', '_').replace('\\', '_') + '.' + self.environment.get_object_suffix()
 
-    def determine_ext_objs(self, extobj, proj_dir_to_build_root):
+    def determine_ext_objs(self, target, extobj, proj_dir_to_build_root):
         result = []
         targetdir = self.get_target_private_dir(extobj.target)
         # With unity builds, there's just one object that contains all the
         # sources, and we only support extracting all the objects in this mode,
         # so just return that.
-        if self.environment.coredata.get_builtin_option('unity'):
+        if self.is_unity(target):
             comp = get_compiler_for_source(extobj.target.compilers.values(),
                                            extobj.srclist[0])
-            # The unity object name uses the full absolute path of the source file
-            osrc = os.path.join(self.get_target_private_dir_abs(extobj.target),
-                                self.get_unity_source_filename(extobj.target,
-                                                               comp.get_default_suffix()))
-            objname = self.object_filename_from_source(extobj.target, osrc)
+            # There is a potential conflict here, but it is unlikely that
+            # anyone both enables unity builds and has a file called foo-unity.cpp.
+            osrc = self.get_unity_source_filename(extobj.target,
+                                                  comp.get_default_suffix())
+            objname = self.object_filename_from_source(extobj.target, osrc, True)
             objpath = os.path.join(proj_dir_to_build_root, targetdir, objname)
             return [objpath]
         for osrc in extobj.srclist:
-            objname = self.object_filename_from_source(extobj.target, osrc)
+            objname = self.object_filename_from_source(extobj.target, osrc, False)
             objpath = os.path.join(proj_dir_to_build_root, targetdir, objname)
             result.append(objpath)
         return result
@@ -343,6 +369,8 @@ class Backend:
         # various sources in the order in which they must override each other
         # starting from hard-coded defaults followed by build options and so on.
         commands = CompilerArgs(compiler)
+
+        copt_proxy = OptionOverrideProxy(target.option_overrides, self.environment.coredata.compiler_options)
         # First, the trivial ones that are impossible to override.
         #
         # Add -nostdinc/-nostdinc++ if needed; can't be overriden
@@ -353,19 +381,19 @@ class Backend:
         # we weren't explicitly asked to not emit warnings (for Vala, f.ex)
         if no_warn_args:
             commands += compiler.get_no_warn_args()
-        elif self.environment.coredata.get_builtin_option('buildtype') != 'plain':
-            commands += compiler.get_warn_args(self.environment.coredata.get_builtin_option('warning_level'))
+        elif self.get_option_for_target('buildtype', target) != 'plain':
+            commands += compiler.get_warn_args(self.get_option_for_target('warning_level', target))
         # Add -Werror if werror=true is set in the build options set on the
         # command-line or default_options inside project(). This only sets the
         # action to be done for warnings if/when they are emitted, so it's ok
         # to set it after get_no_warn_args() or get_warn_args().
-        if self.environment.coredata.get_builtin_option('werror'):
+        if self.get_option_for_target('werror', target):
             commands += compiler.get_werror_args()
         # Add compile args for c_* or cpp_* build options set on the
         # command-line or default_options inside project().
-        commands += compiler.get_option_compile_args(self.environment.coredata.compiler_options)
+        commands += compiler.get_option_compile_args(copt_proxy)
         # Add buildtype args: optimization level, debugging, etc.
-        commands += compiler.get_buildtype_args(self.environment.coredata.get_builtin_option('buildtype'))
+        commands += compiler.get_buildtype_args(self.get_option_for_target('buildtype', target))
         # Add compile args added using add_project_arguments()
         commands += self.build.get_project_args(compiler, target.subproject)
         # Add compile args added using add_global_arguments()
@@ -380,9 +408,11 @@ class Backend:
         # Set -fPIC for static libraries by default unless explicitly disabled
         if isinstance(target, build.StaticLibrary) and target.pic:
             commands += compiler.get_pic_args()
-        # Add compile args needed to find external dependencies
-        # Link args are added while generating the link command
-        for dep in target.get_external_deps():
+        # Add compile args needed to find external dependencies. Link args are
+        # added while generating the link command.
+        # NOTE: We must preserve the order in which external deps are
+        # specified, so we reverse the list before iterating over it.
+        for dep in reversed(target.get_external_deps()):
             commands += dep.get_compile_args()
             # Qt needs -fPIC for executables
             # XXX: We should move to -fPIC for all executables
@@ -405,7 +435,7 @@ class Backend:
         for d in deps:
             if not isinstance(d, (build.StaticLibrary, build.SharedLibrary)):
                 raise RuntimeError('Tried to link with a non-library target "%s".' % d.get_basename())
-            if isinstance(compiler, compilers.LLVMDCompiler):
+            if isinstance(compiler, compilers.LLVMDCompiler) or isinstance(compiler, compilers.DmdDCompiler):
                 args += ['-L' + self.get_target_filename_for_linking(d)]
             else:
                 args.append(self.get_target_filename_for_linking(d))
@@ -444,9 +474,9 @@ class Backend:
         for t in tests:
             exe = t.get_exe()
             if isinstance(exe, dependencies.ExternalProgram):
-                fname = exe.fullpath
+                cmd = exe.get_command()
             else:
-                fname = [os.path.join(self.environment.get_build_dir(), self.get_target_filename(t.get_exe()))]
+                cmd = [os.path.join(self.environment.get_build_dir(), self.get_target_filename(t.get_exe()))]
             is_cross = self.environment.is_cross_build() and \
                 self.environment.cross_info.need_cross_compiler() and \
                 self.environment.cross_info.need_exe_wrapper()
@@ -454,7 +484,7 @@ class Backend:
                 exe_wrapper = self.environment.cross_info.config['binaries'].get('exe_wrapper', None)
             else:
                 exe_wrapper = None
-            if mesonlib.is_windows():
+            if mesonlib.is_windows() or mesonlib.is_cygwin():
                 extra_paths = self.determine_windows_extra_paths(exe)
             else:
                 extra_paths = []
@@ -471,7 +501,7 @@ class Backend:
                     cmd_args.append(self.get_target_filename(a))
                 else:
                     raise MesonException('Bad object in test command.')
-            ts = TestSerialisation(t.get_name(), t.suite, fname, is_cross, exe_wrapper,
+            ts = TestSerialisation(t.get_name(), t.suite, cmd, is_cross, exe_wrapper,
                                    t.is_parallel, cmd_args, t.env, t.should_fail,
                                    t.timeout, t.workdir, extra_paths)
             arr.append(ts)
@@ -545,7 +575,7 @@ class Backend:
         return newargs
 
     def get_build_by_default_targets(self):
-        result = {}
+        result = OrderedDict()
         # Get all build and custom targets that must be built by default
         for name, t in self.build.get_targets().items():
             if t.build_by_default or t.install or t.build_always:
@@ -572,11 +602,18 @@ class Backend:
         for t in target.get_generated_sources():
             if not isinstance(t, build.CustomTarget):
                 continue
-            for f in t.output:
+            for f in t.get_outputs():
                 if self.environment.is_library(f):
                     libs.append(os.path.join(self.get_target_dir(t), f))
         return libs
 
+    def is_unity(self, target):
+        optval = self.get_option_for_target('unity', target)
+        if optval == 'on' or (optval == 'subprojects' and target.subproject != ''):
+            return True
+        return False
+
+
     def get_custom_target_sources(self, target):
         '''
         Custom target sources can be of various object types; strings, File,
@@ -602,20 +639,37 @@ class Backend:
             srcs += fname
         return srcs
 
+    def get_custom_target_depend_files(self, target, absolute_paths=False):
+        deps = []
+        for i in target.depend_files:
+            if isinstance(i, mesonlib.File):
+                if absolute_paths:
+                    deps.append(i.absolute_path(self.environment.get_source_dir(),
+                                                self.environment.get_build_dir()))
+                else:
+                    deps.append(i.rel_to_builddir(self.build_to_src))
+            else:
+                if absolute_paths:
+                    deps.append(os.path.join(self.environment.get_build_dir(), i))
+                else:
+                    deps.append(os.path.join(self.build_to_src, i))
+        return deps
+
     def eval_custom_target_command(self, target, absolute_outputs=False):
-        # We only want the outputs to be absolute when using the VS backend
-        if not absolute_outputs:
-            ofilenames = [os.path.join(self.get_target_dir(target), i) for i in target.output]
-        else:
-            ofilenames = [os.path.join(self.environment.get_build_dir(), self.get_target_dir(target), i)
-                          for i in target.output]
-        srcs = self.get_custom_target_sources(target)
+        # We want the outputs to be absolute only when using the VS backend
+        # XXX: Maybe allow the vs backend to use relative paths too?
+        source_root = self.build_to_src
+        build_root = '.'
         outdir = self.get_target_dir(target)
-        # Many external programs fail on empty arguments.
-        if outdir == '':
-            outdir = '.'
-        if target.absolute_paths:
+        if absolute_outputs:
+            source_root = self.environment.get_source_dir()
+            build_root = self.environment.get_source_dir()
             outdir = os.path.join(self.environment.get_build_dir(), outdir)
+        outputs = []
+        for i in target.get_outputs():
+            outputs.append(os.path.join(outdir, i))
+        inputs = self.get_custom_target_sources(target)
+        # Evaluate the command list
         cmd = []
         for i in target.command:
             if isinstance(i, build.Executable):
@@ -631,37 +685,14 @@ class Backend:
                 if target.absolute_paths:
                     i = os.path.join(self.environment.get_build_dir(), i)
             # FIXME: str types are blindly added ignoring 'target.absolute_paths'
+            # because we can't know if they refer to a file or just a string
             elif not isinstance(i, str):
                 err_msg = 'Argument {0} is of unknown type {1}'
                 raise RuntimeError(err_msg.format(str(i), str(type(i))))
-            for (j, src) in enumerate(srcs):
-                i = i.replace('@INPUT%d@' % j, src)
-            for (j, res) in enumerate(ofilenames):
-                i = i.replace('@OUTPUT%d@' % j, res)
-            if '@INPUT@' in i:
-                msg = 'Custom target {} has @INPUT@ in the command, but'.format(target.name)
-                if len(srcs) == 0:
-                    raise MesonException(msg + ' no input files')
-                if i == '@INPUT@':
-                    cmd += srcs
-                    continue
-                else:
-                    if len(srcs) > 1:
-                        raise MesonException(msg + ' more than one input file')
-                    i = i.replace('@INPUT@', srcs[0])
-            elif '@OUTPUT@' in i:
-                msg = 'Custom target {} has @OUTPUT@ in the command, but'.format(target.name)
-                if len(ofilenames) == 0:
-                    raise MesonException(msg + ' no output files')
-                if i == '@OUTPUT@':
-                    cmd += ofilenames
-                    continue
-                else:
-                    if len(ofilenames) > 1:
-                        raise MesonException(msg + ' more than one output file')
-                    i = i.replace('@OUTPUT@', ofilenames[0])
-            elif '@OUTDIR@' in i:
-                i = i.replace('@OUTDIR@', outdir)
+            elif '@SOURCE_ROOT@' in i:
+                i = i.replace('@SOURCE_ROOT@', source_root)
+            elif '@BUILD_ROOT@' in i:
+                i = i.replace('@BUILD_ROOT@', build_root)
             elif '@DEPFILE@' in i:
                 if target.depfile is None:
                     msg = 'Custom target {!r} has @DEPFILE@ but no depfile ' \
@@ -680,10 +711,11 @@ class Backend:
                     lead_dir = ''
                 else:
                     lead_dir = self.environment.get_build_dir()
-                i = i.replace(source,
-                              os.path.join(lead_dir,
-                                           outdir))
+                i = i.replace(source, os.path.join(lead_dir, outdir))
             cmd.append(i)
+        # Substitute the rest of the template strings
+        values = mesonlib.get_filenames_templates_dict(inputs, outputs)
+        cmd = mesonlib.substitute_values(cmd, values)
         # This should not be necessary but removing it breaks
         # building GStreamer on Windows. The underlying issue
         # is problems with quoting backslashes on Windows
@@ -703,12 +735,12 @@ class Backend:
         #
         # https://github.com/mesonbuild/meson/pull/737
         cmd = [i.replace('\\', '/') for i in cmd]
-        return srcs, ofilenames, cmd
+        return inputs, outputs, cmd
 
     def run_postconf_scripts(self):
         env = {'MESON_SOURCE_ROOT': self.environment.get_source_dir(),
                'MESON_BUILD_ROOT': self.environment.get_build_dir(),
-               }
+               'MESONINTROSPECT': get_meson_script(self.environment, 'mesonintrospect')}
         child_env = os.environ.copy()
         child_env.update(env)
 
